import logging
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

# URI for this module
uri = "/iam/access/v8/dynamic_client_migration"
requires_modules = ["mga"]
requires_version = "10.0.0"


def bulk_migration(isamAppliance, definitionName, definitionId, check_mode=False, force=False):
    """
    Dynamic client bulk migration
    :param isamAppliance:
    :param definitionName:
    :param definitionId:
    :param check_mode:
    :param force:
    :return:
    """

    return isamAppliance.invoke_post(
                "Dynamic client bulk migration",
                "{0}".format(uri),
                {
                    'definitionName': definitionName,
                    'definitionId': definitionId

                },
                requires_modules=requires_modules, requires_version=requires_version)

def client_migration(isamAppliance, definitionName, client_id, check_mode=False, force=False):
    """
    Dynamic client bulk migration
    :param isamAppliance:
    :param definitionName:
    :param definitionId:
    :param check_mode:
    :param force:
    :return:
    """

    return isamAppliance.invoke_post(
                "Dynamic client bulk migration",
                "{0}/{1}".format(uri, client_id),
                {
                    'definitionName': definitionName

                },
                requires_modules=requires_modules, requires_version=requires_version)


import logging

logger = logging.getLogger(__name__)


# *** WORK IN PROGRESS - module not done yet

def get_all(isamAppliance, isamUser, object='/', admin_domain='Default'):
    """
    Retrieve a list of objects
    """
    ret_obj = isamAppliance.invoke_post("Retrieve a list of objects",
                                        "/isam/pdadmin/objectlistandshowext/v1", {
                                            "admin_id": isamUser.username,
                                            "admin_pwd": isamUser.password,
                                            "admin_domain": admin_domain,
                                            "object": object
                                        })
    ret_obj['changed'] = False

    return ret_obj


def get(isamAppliance, isamUser, object, admin_domain='Default'):
    """
    Retrieve a specific object
    """
    ret_obj = isamAppliance.invoke_post("Retrieve a specific object",
                                        "/isam/pdadmin/objectshowext/v1", {
                                            "admin_id": isamUser.username,
                                            "admin_pwd": isamUser.password,
                                            "admin_domain": admin_domain,
                                            "object": object
                                        })
    ret_obj['changed'] = False

    return ret_obj


def compare(isamAppliance1, isamAppliance2, isamUser, admin_domain='Default'):
    """
    Compare objects between two appliances
    """
    ret_obj1 = get_all(isamAppliance1, isamUser, '/', admin_domain)
    ret_obj2 = get_all(isamAppliance2, isamUser, '/', admin_domain)

    for obj in ret_obj1['data']:
        ret_obj = get(isamAppliance1, isamUser=isamUser, object=obj['id'], admin_domain=admin_domain)
        obj['script'] = ret_obj['data']['contents']
    for obj in ret_obj2['data']:
        ret_obj = get(isamAppliance2, isamUser=isamUser, object=obj['id'], admin_domain=admin_domain)
        obj['script'] = ret_obj['data']['contents']

    import ibmsecurity.utilities.tools
    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=[])



import json
import requests
from requests.packages.urllib3.exceptions import InsecureRequestWarning
import logging
from .ibmappliance import IBMAppliance
from .ibmappliance import IBMError
from .ibmappliance import IBMFatal
from ibmsecurity.utilities import tools
from io import open

try:
    basestring
except NameError:
    basestring = (str, bytes)


class ISAMAppliance(IBMAppliance):
    def __init__(self, hostname, user, lmi_port=443, cert=None):
        self.logger = logging.getLogger(__name__)
        self.logger.debug('Creating an ISAMAppliance')
        if isinstance(lmi_port, basestring):
            self.lmi_port = int(lmi_port)
        else:
            self.lmi_port = lmi_port
        self.session = requests.session()
        if cert is None:
            self.logger.debug('Cert object is None, using BA Auth with userid/password.')
            self.session.auth = (user.username, user.password)
        else:
            self.logger.debug('Using cert based auth, since cert object is not None.')
            self.session.cert = cert
        IBMAppliance.__init__(self, hostname, user)

    def _url(self, uri):
        # Build up the URL
        url = "https://" + self.hostname + ":" + str(self.lmi_port) + uri
        self.logger.debug("Issuing request to: " + url)

        return url

    def _log_desc(self, description):
        if description != "":
            self.logger.info('*** ' + description + ' ***')

    def _suppress_ssl_warning(self):
        # Disable https warning because of non-standard certs on appliance
        try:
            self.logger.debug("Suppressing SSL Warnings.")
            requests.packages.urllib3.disable_warnings(InsecureRequestWarning)
        except AttributeError:
            self.logger.warning("load requests.packages.urllib3.disable_warnings() failed")

    def _process_response(self, return_obj, http_response, ignore_error):

        return_obj['rc'] = http_response.status_code

        # Examine the response.
        if (http_response.status_code == 403):
            self.logger.error("  Request failed: ")
            self.logger.error("     status code: {0}".format(http_response.status_code))
            if http_response.text != "":
                self.logger.error("     text: " + http_response.text)
            # Unconditionally raise exception to abort execution
            raise IBMFatal("HTTP Return code: {0}".format(http_response.status_code), http_response.text)
        elif (
                http_response.status_code != 200 and http_response.status_code != 204 and http_response.status_code != 201):
            self.logger.error("  Request failed: ")
            self.logger.error("     status code: {0}".format(http_response.status_code))
            if http_response.text != "":
                self.logger.error("     text: " + http_response.text)
            if not ignore_error:
                raise IBMError("HTTP Return code: {0}".format(http_response.status_code), http_response.text)
            return_obj['changed'] = False  # force changed to be False as there is an error
        else:
            return_obj['rc'] = 0

        # Handle if there was json on input but response was not in json format
        try:
            json_data = json.loads(http_response.text)
            return_obj['data'] = json_data
        except ValueError:
            return_obj['data'] = http_response.content.decode("utf-8")
            return

        self.logger.debug("Status Code: {0}".format(http_response.status_code))
        if http_response.text != "":
            self.logger.debug("Text: " + http_response.content.decode("utf-8"))

        for key in http_response.headers:
            if key == 'g-type':
                if http_response.headers[key] == 'application/octet-stream; charset=UTF-8':
                    json_data = {}
                    return_obj.data = http_response.content.decode("utf-8")
                    return

        if http_response.text == "":
            json_data = {}
        else:
            json_data = json.loads(http_response.text)

        return_obj['data'] = json_data

    def _process_connection_error(self, ignore_error, return_obj):
        if not ignore_error:
            self.logger.critical("Failed to connect to server.")
            raise IBMError("HTTP Return code: 502", "Failed to connect to server")
        else:
            self.logger.debug("Failed to connect to server.")
            return_obj['rc'] = 502

    def _process_warnings(self, uri, requires_modules, requires_version, requires_model, warnings=[]):
        # flag to indicate if processing needs to return and not continue
        return_call = False
        self.logger.debug("Checking for deployment model {0}.".format(requires_model))
        if requires_model is not None and 'model' in self.facts and self.facts['model'] is not None:
            if self.facts['model'] != requires_model:
                return_call = True
                warnings.append(
                    "API invoked requires model: {0}, appliance is of deployment model: {1}.".format(
                        requires_model, self.facts['model']))

        self.logger.debug("Checking for minimum version: {0}.".format(requires_version))
        if requires_version is not None and 'version' in self.facts and self.facts['version'] is not None:
            if tools.version_compare(self.facts['version'], requires_version) < 0:
                return_call = True
                warnings.append(
                    "API invoked requires minimum version: {0}, appliance is of lower version: {1}.".format(
                        requires_version, self.facts['version']))
        # Detecting modules from uri if none is provided
        if requires_modules is None and not requires_modules:
            if uri.startswith("/wga"):
                requires_modules = ['wga']
                self.logger.debug("Detected module: {0} from uri: {1}.".format(requires_modules, uri))
            elif uri.startswith("/mga"):
                requires_modules = ['mga']
                self.logger.debug("Detected module: {0} from uri: {1}.".format(requires_modules, uri))

        self.logger.debug("Checking for one of required modules: {0}.".format(requires_modules))
        if requires_modules is not None and requires_modules:
            if 'activations' in self.facts and self.facts['activations']:
                # Find intersection of the two lists
                iactive = [ia for ia in self.facts['activations'] if ia in requires_modules]
                if not iactive:
                    return_call = True
                    warnings.append(
                        "API invoked requires one of modules: {0}, appliance has these modules active: {1}.".format(
                            requires_modules, self.facts['activations']))
                else:
                    self.logger.info("Modules satisfying requirement: {0}".format(iactive))
            else:
                return_call = True
                warnings.append("API invoked requires module: {0}, appliance has no modules active.".format(
                    requires_modules))

        self.logger.debug("Warnings: {0}".format(warnings))
        return warnings, return_call

    def invoke_post_files(self, description, uri, fileinfo, data, ignore_error=False, requires_modules=None,
                          requires_version=None, warnings=[], json_response=True, data_as_files=False,
                          requires_model=None):
        """
        Send multipart/form-data upload file request to the appliance.
        """
        self._log_desc(description=description)

        warnings, return_call = self._process_warnings(uri=uri, requires_modules=requires_modules,
                                                       requires_version=requires_version,
                                                       warnings=warnings, requires_model=requires_model)
        return_obj = self.create_return_object(warnings=warnings)
        if return_call:
            return return_obj

        # Build up the URL and header information.
        if json_response:
            headers = {
                'Accept': 'application/json,text/html,application/xhtml+xml,application/xml'
            }
        else:
            headers = {
                'Accept': 'text/html,application/xhtml+xml,application/xml'
            }
        self.logger.debug("Headers are: {0}".format(headers))

        if data_as_files is False:
            files = list()
            for file2post in fileinfo:
                files.append((file2post['file_formfield'],
                              (tools.path_leaf(file2post['filename']), open(file2post['filename'], 'rb'),
                               file2post['mimetype'])))
        else:
            files = data

        self._suppress_ssl_warning()

        try:
            if data_as_files is False:
                r = self.session.post(url=self._url(uri=uri), data=data, files=files, verify=False, headers=headers)
            else:
                r = self.session.post(url=self._url(uri=uri), files=files, verify=False, headers=headers)
            return_obj['changed'] = True  # POST of file would be a change
            self._process_response(return_obj=return_obj, http_response=r, ignore_error=ignore_error)

        except requests.exceptions.ConnectionError:
            if not ignore_error:
                self.logger.critical("Failed to connect to server.")
                raise IBMError("HTTP Return code: 502", "Failed to connect to server")
            else:
                self.logger.debug("Failed to connect to server.")
                return_obj.rc = 502

        return return_obj

    def invoke_put_files(self, description, uri, fileinfo, data, ignore_error=False, requires_modules=None,
                         requires_version=None, warnings=[], requires_model=None):
        """
        Send multipart/form-data upload file request to the appliance.
        """
        self._log_desc(description=description)

        warnings, return_call = self._process_warnings(uri=uri, requires_modules=requires_modules,
                                                       requires_version=requires_version, requires_model=requires_model,
                                                       warnings=warnings)
        return_obj = self.create_return_object(warnings=warnings)
        if return_call:
            return return_obj

        # Build up the URL and header information.
        headers = {
            'Accept': 'application/json,text/html,application/xhtml+xml,application/xml'
        }
        self.logger.debug("Headers are: {0}".format(headers))

        files = list()

        for file2post in fileinfo:
            files.append((file2post['file_formfield'],
                          (file2post['filename'], open(file2post['filename'], 'rb'), file2post['mimetype'])))

        self._suppress_ssl_warning()

        try:
            r = self.session.put(url=self._url(uri=uri), data=data, files=files, verify=False, headers=headers)
            return_obj['changed'] = True  # POST of file would be a change
            self._process_response(return_obj=return_obj, http_response=r, ignore_error=ignore_error)

        except requests.exceptions.ConnectionError:
            if not ignore_error:
                self.logger.critical("Failed to connect to server.")
                raise IBMError("HTTP Return code: 502", "Failed to connect to server")
            else:
                self.logger.debug("Failed to connect to server.")
                return_obj.rc = 502

        return return_obj

    def invoke_get_file(self, description, uri, filename, no_headers=False, ignore_error=False, requires_modules=None,
                        requires_version=None, warnings=[], requires_model=None):
        """
        Invoke a GET request and download the response data to a file
        """
        self._log_desc(description=description)

        warnings, return_call = self._process_warnings(uri=uri, requires_modules=requires_modules,
                                                       requires_version=requires_version, requires_model=requires_model,
                                                       warnings=warnings)
        return_obj = self.create_return_object(warnings=warnings)
        if return_call:
            return return_obj

        # In some cases passing a header causes response to come back as JSON
        if no_headers is True:
            headers = {}
        else:
            headers = {
                'Accept': 'application/json,application/octet-stream'
            }
        self.logger.debug("Headers are: {0}".format(headers))

        self._suppress_ssl_warning()

        try:
            r = self.session.get(url=self._url(uri=uri), verify=False, stream=True, headers=headers)

            if (r.status_code != 200 and r.status_code != 204 and r.status_code != 201):
                self.logger.error("  Request failed: ")
                self.logger.error("     status code: {0}".format(r.status_code))
                if r.text != "":
                    self.logger.error("     text: " + r.text)
                if not ignore_error:
                    raise IBMError("HTTP Return code: {0}".format(r.status_code), r.text)
                else:
                    return_obj['rc'] = r.status_code
                    return_obj['data'] = {'msg': 'Unable to extract contents to file!'}
            else:
                with open(filename, 'wb') as f:
                    for chunk in r.iter_content(chunk_size=1024):
                        if chunk:  # filter out keep-alive new chunks
                            f.write(chunk)
                return_obj['rc'] = 0
                return_obj['data'] = {'msg': 'Contents extracted to file: ' + filename}

        except requests.exceptions.ConnectionError:
            self._process_connection_error(ignore_error=ignore_error, return_obj=return_obj)

        except IOError:
            if not ignore_error:
                self.logger.critical("Failed to write to file: " + filename)
                raise IBMError("HTTP Return code: 999", "Failed to write to file: " + filename)
            else:
                self.logger.debug("Failed to write to file: " + filename)
                return_obj['rc'] = 999

        return return_obj

    def _invoke_request(self, func, description, uri, ignore_error, data={}, requires_modules=None,
                        requires_version=None, warnings=[], requires_model=None):
        """
        Send a request to the LMI.  This function is private and should not be
        used directly.  The invoke_get/invoke_put/etc functions should be used instead.
        """
        self._log_desc(description=description)

        warnings, return_call = self._process_warnings(uri=uri, requires_modules=requires_modules,
                                                       requires_version=requires_version, requires_model=requires_model,
                                                       warnings=warnings)
        return_obj = self.create_return_object(warnings=warnings)
        if return_call:
            return return_obj

        # There maybe some cases when header should be blank (not json)
        headers = {
            'Accept': 'application/json',
            'Content-type': 'application/json'
        }
        self.logger.debug("Headers are: {0}".format(headers))

        # Process the input data into JSON
        json_data = json.dumps(data)

        self.logger.debug("Input Data: " + json_data)

        self._suppress_ssl_warning()

        try:
            if func == self.session.get or func == self.session.delete:

                if data != {}:
                    r = func(url=self._url(uri), data=json_data, verify=False, headers=headers)
                else:
                    r = func(url=self._url(uri), verify=False, headers=headers)
            else:
                r = func(url=self._url(uri), data=json_data,
                         verify=False, headers=headers)

            if func != self.session.get:
                return_obj['changed'] = True  # Anything but GET should result in change

            self._process_response(return_obj=return_obj, http_response=r, ignore_error=ignore_error)

        except requests.exceptions.ConnectionError:
            self._process_connection_error(ignore_error=ignore_error, return_obj=return_obj)

        return return_obj

    def _invoke_request_with_headers(self, func, description, uri, ignore_error, headers, data={},
                                     requires_modules=None, requires_version=None, warnings=[], requires_model=None):
        """
        Send a request to the LMI.  This function is private and should not be
        used directly.  The invoke_get/invoke_put/etc functions should be used instead.
        """
        self._log_desc(description=description)

        warnings, return_call = self._process_warnings(uri=uri, requires_modules=requires_modules,
                                                       requires_version=requires_version, requires_model=requires_model,
                                                       warnings=warnings)
        return_obj = self.create_return_object(warnings=warnings)
        if return_call:
            return return_obj

        self.logger.debug("Headers are: {0}".format(headers))

        # Process the input data into JSON
        json_data = json.dumps(data)

        self.logger.debug("Input Data: " + json_data)

        self._suppress_ssl_warning()

        try:
            if func == self.session.get or func == self.session.delete:

                if data != {}:
                    r = func(url=self._url(uri), data=json_data, verify=False, headers=headers)
                else:
                    r = func(url=self._url(uri), verify=False, headers=headers)
            else:
                r = func(url=self._url(uri), data=json_data,
                         verify=False, headers=headers)

            if func != self.session.get:
                return_obj['changed'] = True  # Anything but GET should result in change

            self._process_response(return_obj=return_obj, http_response=r, ignore_error=ignore_error)

        except requests.exceptions.ConnectionError:
            self._process_connection_error(ignore_error=ignore_error, return_obj=return_obj)

        return return_obj

    def invoke_put(self, description, uri, data, ignore_error=False, requires_modules=None, requires_version=None,
                   warnings=[], requires_model=None):
        """
        Send a PUT request to the LMI.
        """

        self._log_request("PUT", uri, description)
        response = self._invoke_request(self.session.put, description, uri,
                                        ignore_error, data,
                                        requires_modules=requires_modules, requires_version=requires_version,
                                        requires_model=requires_model, warnings=warnings)
        return response

    def invoke_post(self, description, uri, data, ignore_error=False, requires_modules=None, requires_version=None,
                    warnings=[], requires_model=None):
        """
        Send a POST request to the LMI.
        """

        self._log_request("POST", uri, description)
        response = self._invoke_request(self.session.post, description, uri,
                                        ignore_error, data,
                                        requires_modules=requires_modules, requires_version=requires_version,
                                        requires_model=requires_model,
                                        warnings=warnings)
        return response

    def invoke_post_snapshot_id(self, description, uri, data, ignore_error=False, requires_modules=None,
                                requires_version=None, warnings=[], requires_model=None):
        """
        Send a POST request to the LMI.  Snapshot id is part of the uri.
        Requires different headers to normal post.
        """

        self._log_request("POST", uri, description)
        warnings, return_call = self._process_warnings(uri=uri, requires_modules=requires_modules,
                                                       requires_version=requires_version, requires_model=requires_model,
                                                       warnings=warnings)

        return_obj = self.create_return_object(warnings=warnings)
        if return_call:
            return return_obj

        headers = {
            'Accept': '*/*',
            'Accept-Encoding': 'gzip, deflate, br',
            'Content-Type': 'application/x-www-form-urlencoded'
        }
        self.logger.debug("Headers are: {0}".format(headers))

        self._suppress_ssl_warning()

        try:
            r = self.session.post(url=self._url(uri=uri), data=data, verify=False, headers=headers)
            return_obj['changed'] = False  # POST of snapshot id would not be a change
            self._process_response(return_obj=return_obj, http_response=r, ignore_error=ignore_error)

        except requests.exceptions.ConnectionError:
            if not ignore_error:
                self.logger.critical("Failed to connect to server.")
                raise IBMError("HTTP Return code: 502", "Failed to connect to server")
            else:
                self.logger.debug("Failed to connect to server.")
                return_obj.rc = 502

        return return_obj

    def invoke_get(self, description, uri, ignore_error=False, requires_modules=None, requires_version=None,
                   warnings=[], requires_model=None):
        """
        Send a GET request to the LMI.
        """
        self._log_request("GET", uri, description)

        response = self._invoke_request(self.session.get, description, uri,
                                        ignore_error, requires_modules=requires_modules,
                                        requires_version=requires_version, requires_model=requires_model,
                                        warnings=warnings)
        self._log_response(response)
        return response

    def invoke_get_with_headers(self, description, uri, headers, ignore_error=False, requires_modules=None,
                                requires_version=None,
                                warnings=[], requires_model=None):
        """
        Send a GET request to the LMI with passed in headers.
        """
        self._log_request("GET", uri, description)

        response = self._invoke_request_with_headers(self.session.get, description, uri,
                                                     ignore_error, headers=headers, requires_modules=requires_modules,
                                                     requires_version=requires_version, requires_model=requires_model,
                                                     warnings=warnings)
        self._log_response(response)
        return response

    def invoke_delete(self, description, uri, data={}, ignore_error=False, requires_modules=None, requires_version=None,
                      warnings=[], requires_model=None):
        """
        Send a DELETE request to the LMI.
        """
        self._log_request("DELETE", uri, description)
        if data != {}:
            self.logger.info("Input Data:{0}".format(data))
            response = self._invoke_request(self.session.delete, description, uri, ignore_error, data=data,
                                            requires_modules=requires_modules, requires_version=requires_version,
                                            requires_model=requires_model,
                                            warnings=warnings)
        else:
            response = self._invoke_request(self.session.delete, description, uri, ignore_error,
                                            requires_modules=requires_modules, requires_version=requires_version,
                                            requires_model=requires_model,
                                            warnings=warnings)
        self._log_response(response)
        return response

    def invoke_request(self, description, method, uri, filename=None, ignore_error=False, requires_modules=None,
                       requires_version=None,
                       warnings=[], requires_model=None, **kwargs):
        """
        parse and send a appropriate request to the appliance.
        """
        self._log_desc(description=description)

        warnings, return_call = self._process_warnings(uri=uri, requires_modules=requires_modules,
                                                       requires_version=requires_version, requires_model=requires_model,
                                                       warnings=warnings)
        return_obj = self.create_return_object(warnings=warnings)
        if return_call:
            return return_obj

        args = {}

        for key, value in kwargs.items():
            if key == 'json' and value != {}:
                json_data = json.dumps(value)
                self.logger.debug("Input json Data: " + json_data)
                args['json'] = json_data
            elif key == 'data':
                try:
                    json.loads(value)
                    self.logger.debug("Input Data: " + value)
                    args['data'] = value
                except ValueError:
                    self.logger.debug("Input Data: " + value)
                    args['data'] = value
            else:
                args[key] = value

        self._suppress_ssl_warning()

        try:
            streaminargs = False
            r = self.session.request(method, url=self._url(uri), verify=False, **args)
            # check for stream=True
            if "stream" in args and args["stream"] == True:
                streaminargs = True
                if filename == None:
                    return_obj['warnings'] = return_obj['warnings'].append(
                        "filename is missing, for stream=True, filename needs to be non null")
                    return return_obj
                # else stream content to file
                else:
                    if (r.status_code != 200 and r.status_code != 204 and r.status_code != 201):
                        self.logger.error("  Request failed: ")
                        self.logger.error("     status code: {0}".format(r.status_code))
                        if r.text != "":
                            self.logger.error("     text: " + r.text)
                        if not ignore_error:
                            raise IBMError("HTTP Return code: {0}".format(r.status_code), r.text)
                        else:
                            return_obj['rc'] = r.status_code
                            return_obj['data'] = {'msg': 'Unable to extract contents to file!'}
                    else:
                        with open(filename, 'wb') as f:
                            for chunk in r.iter_content(chunk_size=1024):
                                if chunk:  # filter out keep-alive new chunks
                                    f.write(chunk)
                        return_obj['rc'] = 0
                        return_obj['data'] = {'msg': 'Contents extracted to file: ' + filename}

            if method == "get" or (method == "post" and streaminargs == True):
                return_obj['changed'] = False
            else:
                return_obj['changed'] = True  # Anything but GET or a POST with stream=True set should result in change

            if streaminargs == False:
                self._process_response(return_obj=return_obj, http_response=r, ignore_error=ignore_error)

        except requests.exceptions.ConnectionError:
            self._process_connection_error(ignore_error=ignore_error, return_obj=return_obj)

        return return_obj

    def get_facts(self):
        """
        Get facts about the appliance
        """
        # Fact collection will abort on any exception
        try:
            self.get_version()

            # Check if appliance is setup before collecting Activation information
            import ibmsecurity.isam.base.setup_complete
            ret_obj = ibmsecurity.isam.base.setup_complete.get(self)
            if ret_obj['data'].get('configured') is True:
                self.get_activations()
        # Be sure to let fatal error unconditionally percolate up the stack
        except IBMFatal:
            raise
        # Exceptions like those connection related will be ignored
        except:
            pass

    def get_version(self):
        """
        Get appliance version (versions API or active partition)

        When firmware are installed or partition are changed, then this value is updated
        """
        self.facts['version'] = None
        import ibmsecurity.isam.base.version
        import ibmsecurity.isam.base.firmware

        try:
            ret_obj = ibmsecurity.isam.base.version.get(self)
            self.facts['version'] = ret_obj['data']['firmware_version']

            if tools.version_compare(self.facts['version'], '9.0.3.0') > 0:
                if 'deployment_model' in ret_obj['data']:
                    self.facts['model'] = ret_obj['data']['deployment_model']

                if 'product_name' in ret_obj['data']:
                    self.facts['product_name'] = ret_obj['data']['product_name']

                if 'product_description' in ret_obj['data']:
                    self.facts['product_description'] = ret_obj['data']['product_description']

                if 'firmware_build' in ret_obj['data']:
                    self.facts['firmware_build'] = ret_obj['data']['firmware_build']

                if 'firmware_label' in ret_obj['data']:
                    self.facts['firmware_label'] = ret_obj['data']['firmware_label']

        # Be sure to let fatal error unconditionally percolate up the stack
        except IBMFatal:
            raise
        except IBMError:
            try:
                ret_obj = ibmsecurity.isam.base.firmware.get(self)
                for partition in ret_obj['data']:
                    if partition['active'] is True:
                        ver = partition['firmware_version'].split(' ')
                        self.facts['version'] = ver[-1]
                self.facts['model'] = "Appliance"
            except:
                pass
        return

    def get_activations(self):
        """
        Get  appliance activations

        When new modules are activated or old ones de-activated this value is updated.
        """
        self.facts['activations'] = []
        import ibmsecurity.isam.base.activation

        ret_obj = ibmsecurity.isam.base.activation.get_all(self)
        for activation in ret_obj['data']:
            if activation['enabled'] == 'True':
                self.facts['activations'].append(activation['id'])

    def _log_request(self, method, url, desc):
        self.logger.debug("Request: %s %s desc=%s", method, url, desc)

    def _log_response(self, response):
        if response:
            self.logger.debug("Response: %d", response.get('rc'))
            # self.logger.debug("Response: %i %i warnings:%s",
            #                     response.get('rc'),
            #                     response.get('status_code'),
            #                     response.get('warnings'))
        else:
            self.logger.debug("Response: None")



import logging
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)


def get_all(isamAppliance, check_mode=False, force=False):
    """
    Retrieving the complete list of SCIM configuration settings
    """
    return isamAppliance.invoke_get("Retrieving the complete list of SCIM configuration settings",
                                    "/mga/scim/configuration")


def get_user_profile(isamAppliance, check_mode=False, force=False):
    """
    Retrieve configuration of SCIM user profile settings
    """
    ret_obj = isamAppliance.invoke_get("Retrieve configuration of SCIM user profile settings",
                                       "/mga/scim/configuration/urn:ietf:params:scim:schemas:core:2.0:User")
    return ret_obj['data']['urn:ietf:params:scim:schemas:core:2.0:User']


def get_isam_user(isamAppliance, check_mode=False, force=False):
    """
    Retrieve configuration of SCIM ISAM user settings
    """
    ret_obj = isamAppliance.invoke_get("Retrieve configuration of SCIM ISAM user settings",
                                       "/mga/scim/configuration/urn:ietf:params:scim:schemas:extension:isam:1.0:User")
    return ret_obj['data']['urn:ietf:params:scim:schemas:extension:isam:1.0:User']


def update_user_profile(isamAppliance, ldap_connection, user_suffix, search_suffix, check_mode=False, force=False):
    """
    Update SCIM user profile settings
    """
    ret_obj = get_user_profile(isamAppliance)
    del ret_obj['ldap_connection']
    del ret_obj['user_suffix']
    del ret_obj['search_suffix']

    ret_obj['ldap_connection'] = ldap_connection
    ret_obj['user_suffix'] = user_suffix
    ret_obj['search_suffix'] = search_suffix
    return isamAppliance.invoke_put(
        "Update SCIM user profile settings",
        "/mga/scim/configuration/urn:ietf:params:scim:schemas:core:2.0:User",
        ret_obj)


def update_isam_user(isamAppliance, isam_domain, update_native_users, ldap_connection, check_mode=False, force=False):
    """
    Update SCIM ISAM user settings
    """
    ret_obj = get_isam_user(isamAppliance)
    del ret_obj['isam_domain']
    del ret_obj['update_native_users']
    del ret_obj['ldap_connection']

    ret_obj['isam_domain'] = isam_domain
    ret_obj['update_native_users'] = update_native_users
    ret_obj['ldap_connection'] = ldap_connection
    return isamAppliance.invoke_put(
        "Update SCIM ISAM user settings",
        "/mga/scim/configuration/urn:ietf:params:scim:schemas:extension:isam:1.0:User",
        ret_obj)


def set_all(isamAppliance, scim_configuration, check_mode=False, force=False):
    """
    Update entire SCIM settings
    """
    if scim_configuration is None or scim_configuration == '':
        return isamAppliance.create_return_object(
            warnings="Need to pass content for scim configuration")
    else:
        if force is True or _check(isamAppliance, scim_configuration) is False:
            if check_mode is True:
                return isamAppliance.create_return_object(changed=True)
            else:
                return isamAppliance.invoke_put(
                    "Update SCIM settings",
                    "/mga/scim/configuration",
                    scim_configuration)

    return isamAppliance.create_return_object()


def _check(isamAppliance, scim_configuration):
    """
    Check if scim configuration is identical with server
    """
    ret_obj = get_all(isamAppliance)
    logger.debug("Comparing server scim configuration with desired configuration.")
    cur_sorted_json = tools.json_sort(ret_obj['data'])
    logger.debug("Server JSON : {0}".format(cur_sorted_json))
    given_sorted_json = tools.json_sort(scim_configuration)
    logger.debug("Desired JSON: {0}".format(given_sorted_json))
    if cur_sorted_json != given_sorted_json:
        return False
        logger.debug("Changes detected!")
    else:
        logger.debug("Server configuration is identical with desired configuration. No change necessary.")
        return True



import logging
from ibmsecurity.utilities import tools

try:
    basestring
except NameError:
    basestring = (str, bytes)

logger = logging.getLogger(__name__)

# URI for this module
uri = "/wga/reverseproxy"
requires_modules = ["wga"]
requires_version = None


def get_all(isamAppliance, reverseproxy_id, check_mode=False, force=False):
    """
    Retrieving a list of standard and virtual junctions

    :param isamAppliance:
    :param reverseproxy_id:
    :param check_mode:
    :param force:
    :return:
    """
    return isamAppliance.invoke_get("Retrieving a list of standard and virtual junctions",
                                    "{0}/{1}/junctions".format(uri, reverseproxy_id),
                                    requires_modules=requires_modules,
                                    requires_version=requires_version)


def get(isamAppliance, reverseproxy_id, junctionname, check_mode=False, force=False):
    """
    Retrieving the parameters for a single standard or virtual junction

    :param isamAppliance:
    :param reverseproxy_id:
    :param junctionname:
    :param check_mode:
    :param force:
    :return:
    """
    ret_obj = isamAppliance.invoke_get("Retrieving the parameters for a single standard or virtual junction",
                                       "{0}/{1}/junctions?junctions_id={2}".format(uri, reverseproxy_id,
                                                                                   junctionname),
                                       requires_modules=requires_modules,
                                       requires_version=requires_version)
    # servers are provided as a single string, here we parse it out into a list + dict
    servers = []
    if tools.version_compare(isamAppliance.facts["version"], "9.0.1.0") > 0:
        srv_separator = '#'
    else:
        srv_separator = '&'
    logger.debug("Server Separator being used: {0}".format(srv_separator))
    srvs = ret_obj['data']['servers'].split(srv_separator)
    logger.debug("Servers in raw string: {0}".format(ret_obj['data']['servers']))
    logger.debug("Number of servers in junction: {0}".format(len(srvs)))
    for srv in srvs:
        logger.debug("Parsing Server: {0}".format(srv))
        server = {}
        for s in srv.split(';'):
            if s != '':
                kv = s.split('!')
                server[kv[0]] = kv[1]
        servers.append(server)

    ret_obj['data']['servers'] = servers

    return ret_obj


def _check(isamAppliance, reverseproxy_id, junctionname):
    ret_obj = get_all(isamAppliance, reverseproxy_id)

    for jct in ret_obj['data']:
        if jct['id'] == junctionname:
            return True

    return False


def add(isamAppliance, reverseproxy_id, junction_point, server_hostname, server_port, junction_type="tcp",
        virtual_hostname=None, server_dn=None, query_contents=None, stateful_junction=None, case_sensitive_url=None,
        windows_style_url=None, https_port=None, http_port=None, proxy_hostname=None, proxy_port=None,
        sms_environment=None, vhost_label=None, junction_hard_limit=None, junction_soft_limit=None,
        basic_auth_mode=None, tfim_sso=None, remote_http_header=None, preserve_cookie=None, cookie_include_path=None,
        transparent_path_junction=None, mutual_auth=None, insert_session_cookies=None, request_encoding=None,
        enable_basic_auth=None, key_label=None, gso_resource_group=None, junction_cookie_javascript_block=None,
        client_ip_http=None, version_two_cookies=None, ltpa_keyfile=None, authz_rules=None, fsso_config_file=None,
        username=None, password=None, server_uuid=None, local_ip=None, ltpa_keyfile_password=None,
        delegation_support=None, scripting_support=None, insert_ltpa_cookies=None, check_mode=False, force=False,
        http2_junction=None, http2_proxy=None, sni_name=None, description=None, warnings=[]):
    """
    Creating a standard or virtual junction

    :param isamAppliance:
    :param reverseproxy_id:
    :param junction_point:
    :param server_hostname:
    :param server_port:
    :param junction_type:
    :param virtual_hostname:
    :param server_dn:
    :param query_contents:
    :param stateful_junction:
    :param case_sensitive_url:
    :param windows_style_url:
    :param https_port:
    :param http_port:
    :param proxy_hostname:
    :param proxy_port:
    :param sms_environment:
    :param vhost_label:
    :param junction_hard_limit:
    :param junction_soft_limit:
    :param basic_auth_mode:
    :param tfim_sso:
    :param remote_http_header:
    :param preserve_cookie:
    :param cookie_include_path:
    :param transparent_path_junction:
    :param mutual_auth:
    :param insert_session_cookies:
    :param request_encoding:
    :param enable_basic_auth:
    :param key_label:
    :param gso_resource_group:
    :param junction_cookie_javascript_block:
    :param client_ip_http:
    :param version_two_cookies:
    :param ltpa_keyfile:
    :param authz_rules:
    :param fsso_config_file:
    :param username:
    :param password:
    :param server_uuid:
    :param local_ip:
    :param ltpa_keyfile_password:
    :param delegation_support:
    :param scripting_support:
    :param insert_ltpa_cookies:
    :param check_mode:
    :param force:
    :param http2_junction:
    :param http2_proxy:
    :param sni_name:
    :param description:
    :return:
    """
    if force is True or _check(isamAppliance, reverseproxy_id, junction_point) is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            # Create a simple json with just the main junction attributes
            jct_json = {
                "junction_point": junction_point,
                "junction_type": junction_type,
                "server_hostname": server_hostname,
                "server_port": server_port,
                "force": force
            }

            # Add attributes that have been supplied... otherwise skip them.
            if junction_hard_limit is not None:
                jct_json["junction_hard_limit"] = junction_hard_limit
            if junction_soft_limit is not None:
                jct_json["junction_soft_limit"] = junction_soft_limit
            if basic_auth_mode is not None:
                jct_json["basic_auth_mode"] = basic_auth_mode
            if tfim_sso is not None:
                jct_json["tfim_sso"] = tfim_sso
            if remote_http_header is not None:
                jct_json["remote_http_header"] = remote_http_header
            if stateful_junction is not None:
                jct_json["stateful_junction"] = stateful_junction
            if preserve_cookie is not None:
                jct_json["preserve_cookie"] = preserve_cookie
            if cookie_include_path is not None:
                jct_json["cookie_include_path"] = cookie_include_path
            if transparent_path_junction is not None:
                jct_json["transparent_path_junction"] = transparent_path_junction
            if mutual_auth is not None:
                jct_json["mutual_auth"] = mutual_auth
            if insert_ltpa_cookies is not None:
                jct_json["insert_ltpa_cookies"] = insert_ltpa_cookies
            if insert_session_cookies is not None:
                jct_json["insert_session_cookies"] = insert_session_cookies
            if request_encoding is not None:
                jct_json["request_encoding"] = request_encoding
            if enable_basic_auth is not None:
                jct_json["enable_basic_auth"] = enable_basic_auth
            if key_label is not None:
                jct_json["key_label"] = key_label
            if gso_resource_group is not None:
                jct_json["gso_resource_group"] = gso_resource_group
            if junction_cookie_javascript_block is not None and junction_cookie_javascript_block != '':
                jct_json["junction_cookie_javascript_block"] = junction_cookie_javascript_block
            if client_ip_http is not None:
                jct_json["client_ip_http"] = client_ip_http
            if version_two_cookies is not None:
                jct_json["version_two_cookies"] = version_two_cookies
            if ltpa_keyfile is not None:
                jct_json["ltpa_keyfile"] = ltpa_keyfile
            if authz_rules is not None:
                jct_json["authz_rules"] = authz_rules
            if fsso_config_file is not None:
                jct_json["fsso_config_file"] = fsso_config_file
            if username is not None:
                jct_json["username"] = username
            if password is not None:
                jct_json["password"] = password
            if server_uuid is not None:
                jct_json["server_uuid"] = server_uuid
            if virtual_hostname is not None:
                jct_json["virtual_hostname"] = virtual_hostname
            if server_dn is not None:
                jct_json["server_dn"] = server_dn
            if local_ip is not None:
                jct_json["local_ip"] = local_ip
            if query_contents is not None:
                jct_json["query_contents"] = query_contents
            if case_sensitive_url is not None:
                jct_json["case_sensitive_url"] = case_sensitive_url
            if windows_style_url is not None:
                jct_json["windows_style_url"] = windows_style_url
            if ltpa_keyfile_password is not None:
                jct_json["ltpa_keyfile_password"] = ltpa_keyfile_password
            if https_port is not None:
                jct_json["https_port"] = https_port
            if http_port is not None:
                jct_json["http_port"] = http_port
            if proxy_hostname is not None:
                jct_json["proxy_hostname"] = proxy_hostname
            if proxy_port is not None:
                jct_json["proxy_port"] = proxy_port
            if sms_environment is not None:
                jct_json["sms_environment"] = sms_environment
            if vhost_label is not None:
                jct_json["vhost_label"] = vhost_label
            if delegation_support is not None:
                jct_json["delegation_support"] = delegation_support
            if scripting_support is not None:
                jct_json["scripting_support"] = scripting_support
            if http2_junction is not None:
                if tools.version_compare(isamAppliance.facts["version"], "9.0.4.0") < 0:
                    warnings.append(
                        "Appliance at version: {0}, http2_junction: {1} is not supported. Needs 9.0.4.0 or higher. Ignoring http2_junction for this call.".format(
                            isamAppliance.facts["version"], http2_junction))
                else:
                    jct_json["http2_junction"] = http2_junction
            if http2_proxy is not None:
                if tools.version_compare(isamAppliance.facts["version"], "9.0.4.0") < 0:
                    warnings.append(
                        "Appliance at version: {0}, http2_proxy: {1} is not supported. Needs 9.0.4.0 or higher. Ignoring http2_proxy for this call.".format(
                            isamAppliance.facts["version"], http2_proxy))
                else:
                    jct_json['http2_proxy'] = http2_proxy
            if sni_name is not None:
                if tools.version_compare(isamAppliance.facts["version"], "9.0.4.0") < 0:
                    warnings.append(
                        "Appliance at version: {0}, sni_name: {1} is not supported. Needs 9.0.4.0 or higher. Ignoring sni_name for this call.".format(
                            isamAppliance.facts["version"], sni_name))
                else:
                    jct_json['sni_name'] = sni_name
            if description is not None:
                if tools.version_compare(isamAppliance.facts["version"], "9.0.7.0") < 0:
                    warnings.append(
                        "Appliance at version: {0}, description: {1} is not supported. Needs 9.0.7.0 or higher. Ignoring description for this call.".format(
                            isamAppliance.facts["version"], description))
                else:
                    jct_json['description'] = description

            return isamAppliance.invoke_post(
                "Creating a standard or virtual junction",
                "{0}/{1}/junctions".format(uri, reverseproxy_id), jct_json,
                requires_modules=requires_modules,
                requires_version=requires_version, warnings=warnings)

    return isamAppliance.create_return_object(warnings=warnings)


def delete(isamAppliance, reverseproxy_id, junctionname, check_mode=False, force=False):
    """
    Deleting a standard or virtual junction

    :param isamAppliance:
    :param reverseproxy_id:
    :param junctionname:
    :param check_mode:
    :param force:
    :return:
    """
    if force is True or _check(isamAppliance, reverseproxy_id, junctionname) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Deleting a standard or virtual junction",
                "{0}/{1}/junctions?junctions_id={2}".format(uri, reverseproxy_id, junctionname),
                requires_modules=requires_modules,
                requires_version=requires_version)

    return isamAppliance.create_return_object()


def set(isamAppliance, reverseproxy_id, junction_point, server_hostname, server_port, junction_type="tcp",
        virtual_hostname=None, server_dn=None, query_contents=None, stateful_junction=None, case_sensitive_url=None,
        windows_style_url=None, https_port=None, http_port=None, proxy_hostname=None, proxy_port=None,
        sms_environment=None, vhost_label=None, junction_hard_limit=None, junction_soft_limit=None,
        basic_auth_mode=None, tfim_sso=None, remote_http_header=None, preserve_cookie=None, cookie_include_path=None,
        transparent_path_junction=None, mutual_auth=None, insert_session_cookies=None, request_encoding=None,
        enable_basic_auth=None, key_label=None, gso_resource_group=None, junction_cookie_javascript_block=None,
        client_ip_http=None, version_two_cookies=None, ltpa_keyfile=None, authz_rules=None, fsso_config_file=None,
        username=None, password=None, server_uuid=None, local_ip=None, ltpa_keyfile_password=None,
        delegation_support=None, scripting_support=None, insert_ltpa_cookies=None, check_mode=False, force=False,
        http2_junction=None, http2_proxy=None, sni_name=None, description=None):
    """
    Setting a standard or virtual junction - compares with existing junction and replaces if changes are detected
    TODO: Compare all the parameters in the function - LTPA, BA are some that are not being compared
    """
    warnings = []
    add_required = False
    if force is False:
        # Check if record exists
        logger.debug("Check if the junction exists.")
        if _check(isamAppliance, reverseproxy_id, junction_point) is True:
            logger.debug("Junction exists. Compare junction details.")
            ret_obj = get(isamAppliance, reverseproxy_id=reverseproxy_id, junctionname=junction_point)
            server_found = False
            logger.debug("See if the backend junction server matches any on the junction. Look for just one match.")
            srvs = ret_obj['data']['servers']
            srvs_len = len(srvs)
            for srv in srvs:
                if srv['server_hostname'] == server_hostname and str(srv['server_port']) == str(server_port):
                    logger.debug("Matched a server - {0}.".format(srv))
                    server_found = True
                    server_json = {
                        'server_hostname': server_hostname,
                        'server_port': str(server_port)
                    }
                    if case_sensitive_url is None:
                        server_json['case_sensitive_url'] = 'no'
                    else:
                        server_json['case_sensitive_url'] = case_sensitive_url
                    if http_port is None:
                        server_json['http_port'] = str(server_port)
                    else:
                        server_json['http_port'] = str(http_port)
                    if local_ip is None:
                        server_json['local_ip'] = ''
                    else:
                        server_json['local_ip'] = local_ip
                    if query_contents is None:
                        server_json['query_content_url'] = '/cgi-bin/query_contents'
                    else:
                        server_json['query_content_url'] = query_contents
                    if server_dn is None:
                        server_json['server_dn'] = ''
                    else:
                        server_json['server_dn'] = server_dn
                    if server_uuid is not None:
                        server_json['server_uuid'] = server_uuid
                    else:
                        # Server UUID gets generated if not specified
                        if 'server_uuid' in srv:
                            del srv['server_uuid']
                    if virtual_hostname is None:
                        if 'virtual_junction_hostname' in srv:
                            del srv['virtual_junction_hostname']
                    else:
                        server_json['virtual_junction_hostname'] = virtual_hostname
                    if windows_style_url is None:
                        server_json['windows_style_url'] = 'no'
                    else:
                        server_json['windows_style_url'] = windows_style_url

                    # Delete dynamic data shown when we get junctions details
                    if 'current_requests' in srv:
                        del srv['current_requests']
                    if 'total_requests' in srv:
                        del srv['total_requests']
                    if 'operation_state' in srv:
                        del srv['operation_state']
                    if 'server_state' in srv:
                        del srv['server_state']
                    # Not sure what this attribute is supposed to contain?
                    if 'query_contents' in srv:
                        del srv['query_contents']
                    if tools.json_sort(server_json) != tools.json_sort(srv):
                        logger.debug("Servers are found to be different. See following JSON for difference.")
                        logger.debug("New Server JSON: {0}".format(tools.json_sort(server_json)))
                        logger.debug("Old Server JSON: {0}".format(tools.json_sort(srv)))
                        add_required = True
                    break
            if server_found is False:
                add_required = True
            elif add_required is False:
                exist_jct = ret_obj['data']
                jct_json = {
                    'junction_point': junction_point,
                    'junction_type': junction_type.upper()
                }
                if authz_rules is None:
                    jct_json['authz_rules'] = 'no'
                else:
                    jct_json['authz_rules'] = authz_rules
                if basic_auth_mode is None:
                    jct_json['basic_auth_mode'] = 'filter'
                else:
                    jct_json['basic_auth_mode'] = basic_auth_mode
                if client_ip_http is None or client_ip_http.lower() == 'no':
                    jct_json['client_ip_http'] = 'do not insert'
                elif client_ip_http.lower() == 'yes':
                    jct_json['client_ip_http'] = 'insert'
                else:
                    jct_json['client_ip_http'] = client_ip_http
                if cookie_include_path is None:
                    jct_json['cookie_include_path'] = 'no'
                else:
                    jct_json['cookie_include_path'] = cookie_include_path
                if delegation_support is None:
                    jct_json['delegation_support'] = 'no'
                else:
                    jct_json['delegation_support'] = delegation_support
                if fsso_config_file is None:
                    jct_json['fsso_config_file'] = 'disabled'
                else:
                    jct_json['fsso_config_file'] = fsso_config_file
                if insert_session_cookies is None:
                    jct_json['insert_session_cookies'] = 'no'
                else:
                    jct_json['insert_session_cookies'] = insert_session_cookies
                if junction_hard_limit is None:
                    jct_json['junction_hard_limit'] = '0 - using global value'
                else:
                    jct_json['junction_hard_limit'] = str(junction_hard_limit)
                if junction_soft_limit is None:
                    jct_json['junction_soft_limit'] = '0 - using global value'
                else:
                    jct_json['junction_soft_limit'] = str(junction_soft_limit)
                # We could have a comma delimited set of values - so split them into array
                if junction_cookie_javascript_block is not None and junction_cookie_javascript_block != '':
                    jct_json['junction_cookie_javascript_block'] = junction_cookie_javascript_block.split(',')
                    # Here the list is delimited by space
                    if 'junction_cookie_javascript_block' in exist_jct and exist_jct[
                        'junction_cookie_javascript_block'] is not None:
                        exist_jct['junction_cookie_javascript_block'] = exist_jct[
                            'junction_cookie_javascript_block'].split(' ')
                if mutual_auth is None:
                    jct_json['mutual_auth'] = 'no'
                else:
                    jct_json['mutual_auth'] = mutual_auth
                if preserve_cookie is None:
                    jct_json['preserve_cookie'] = 'no'
                else:
                    jct_json['preserve_cookie'] = preserve_cookie
                if remote_http_header is None or remote_http_header == []:
                    jct_json['remote_http_header'] = 'do not insert'
                elif isinstance(remote_http_header, basestring) and remote_http_header.lower() == 'all':
                    jct_json['remote_http_header'] = ['iv_creds', 'iv_groups', 'iv_user']
                else:
                    jct_json['remote_http_header'] = remote_http_header
                # To allow for multiple header values to be sorted during compare convert retrieved data into array
                if exist_jct['remote_http_header'].startswith('insert - '):
                    exist_jct['remote_http_header'] = (exist_jct['remote_http_header'][9:]).split(' ')
                if request_encoding is None:
                    jct_json['request_encoding'] = 'UTF-8, URI Encoded'
                else:
                    jct_json['request_encoding'] = request_encoding
                if scripting_support is None:
                    jct_json['scripting_support'] = 'no'
                else:
                    jct_json['scripting_support'] = scripting_support
                if stateful_junction is None:
                    jct_json['stateful_junction'] = 'no'
                else:
                    jct_json['stateful_junction'] = stateful_junction
                if tfim_sso is None:
                    jct_json['tfim_sso'] = 'no'
                else:
                    jct_json['tfim_sso'] = tfim_sso
                if transparent_path_junction is None:
                    jct_json['transparent_path_junction'] = 'no'
                else:
                    jct_json['transparent_path_junction'] = transparent_path_junction
                if http2_junction is not None:
                    if tools.version_compare(isamAppliance.facts["version"], "9.0.4.0") < 0:
                        warnings.append(
                            "Appliance at version: {0}, http2_junction: {1} is not supported. Needs 9.0.4.0 or higher. Ignoring http2_junction for this call.".format(
                                isamAppliance.facts["version"], http2_junction))
                        http2_junction = None
                    else:
                        jct_json['http2_junction'] = http2_junction
                if http2_proxy is not None:
                    if tools.version_compare(isamAppliance.facts["version"], "9.0.4.0") < 0:
                        warnings.append(
                            "Appliance at version: {0}, http2_proxy: {1} is not supported. Needs 9.0.4.0 or higher. Ignoring http2_proxy for this call.".format(
                                isamAppliance.facts["version"], http2_proxy))
                        http2_proxy = None
                    else:
                        jct_json['http2_proxy'] = http2_proxy
                if sni_name is not None:
                    if tools.version_compare(isamAppliance.facts["version"], "9.0.4.0") < 0:
                        warnings.append(
                            "Appliance at version: {0}, sni_name: {1} is not supported. Needs 9.0.4.0 or higher. Ignoring sni_name for this call.".format(
                                isamAppliance.facts["version"], sni_name))
                        sni_name = None
                    else:
                        jct_json['sni_name'] = sni_name
                if description is not None:
                    if tools.version_compare(isamAppliance.facts["version"], "9.0.7.0") < 0:
                        warnings.append(
                            "Appliance at version: {0}, description: {1} is not supported. Needs 9.0.7.0 or higher. Ignoring description for this call.".format(
                                isamAppliance.facts["version"], description))
                        description = None
                    else:
                        jct_json['description'] = description

                # TODO: Not sure of how to match following attributes! Need to revisit.
                # TODO: Not all function parameters are being checked - need to add!
                del exist_jct['boolean_rule_header']
                del exist_jct['forms_based_sso']
                del exist_jct['http_header_ident']
                del exist_jct['session_cookie_backend_portal']
                # We are already comparing server details - so remove this from this compare
                del exist_jct['servers']
                # Delete dynamic data shown when we get junctions details
                del exist_jct['active_worker_threads']
                if tools.json_sort(jct_json) != tools.json_sort(exist_jct):
                    logger.debug("Junctions are found to be different. See following JSON for difference.")
                    logger.debug("New Junction JSON: {0}".format(tools.json_sort(jct_json)))
                    logger.debug("Old Junction JSON: {0}".format(tools.json_sort(exist_jct)))
                    add_required = True
            if add_required is True and srvs_len > 1:
                warnings.append(
                    "Junction will replaced. Existing multiple servers #{0} will be overwritten. Please re-add as needed.".format(
                        srvs_len))
        else:
            add_required = True

    if force is True or add_required is True:
        # Junction force add will replace the junction, no need for delete (force set to True as a result)
        return add(isamAppliance=isamAppliance, reverseproxy_id=reverseproxy_id, junction_point=junction_point,
                   server_hostname=server_hostname, server_port=server_port, junction_type=junction_type,
                   virtual_hostname=virtual_hostname, server_dn=server_dn, query_contents=query_contents,
                   stateful_junction=stateful_junction, case_sensitive_url=case_sensitive_url,
                   windows_style_url=windows_style_url, https_port=https_port, http_port=http_port,
                   proxy_hostname=proxy_hostname, proxy_port=proxy_port, sms_environment=sms_environment,
                   vhost_label=vhost_label, junction_hard_limit=junction_hard_limit,
                   junction_soft_limit=junction_soft_limit, basic_auth_mode=basic_auth_mode, tfim_sso=tfim_sso,
                   remote_http_header=remote_http_header, preserve_cookie=preserve_cookie,
                   cookie_include_path=cookie_include_path, transparent_path_junction=transparent_path_junction,
                   mutual_auth=mutual_auth, insert_session_cookies=insert_session_cookies,
                   request_encoding=request_encoding, enable_basic_auth=enable_basic_auth, key_label=key_label,
                   gso_resource_group=gso_resource_group,
                   junction_cookie_javascript_block=junction_cookie_javascript_block, client_ip_http=client_ip_http,
                   version_two_cookies=version_two_cookies, ltpa_keyfile=ltpa_keyfile, authz_rules=authz_rules,
                   fsso_config_file=fsso_config_file, username=username, password=password, server_uuid=server_uuid,
                   local_ip=local_ip, ltpa_keyfile_password=ltpa_keyfile_password,
                   delegation_support=delegation_support, scripting_support=scripting_support,
                   insert_ltpa_cookies=insert_ltpa_cookies, check_mode=check_mode, force=True,
                   http2_junction=http2_junction, http2_proxy=http2_proxy, sni_name=sni_name, description=description, warnings=warnings)

    return isamAppliance.create_return_object()


def compare(isamAppliance1, isamAppliance2, reverseproxy_id, reverseproxy_id2=None):
    """
    Compare list of junctions in a given reverse proxy between 2 appliances
    """
    if reverseproxy_id2 is None or reverseproxy_id2 == '':
        reverseproxy_id2 = reverseproxy_id
    ret_obj1 = get_all(isamAppliance1, reverseproxy_id)
    ret_obj2 = get_all(isamAppliance2, reverseproxy_id2)

    for jct in ret_obj1['data']:
        ret_obj = get(isamAppliance1, reverseproxy_id, jct['id'])
        del ret_obj['data']['active_worker_threads']
        for srv in ret_obj['data']['servers']:
            del srv['current_requests']
            del srv['operation_state']
            del srv['server_state']
            if ret_obj['data']['stateful_junction'] == 'no':
                del srv['server_uuid']
            del srv['total_requests']
        jct['details'] = ret_obj['data']

    for jct in ret_obj2['data']:
        ret_obj = get(isamAppliance2, reverseproxy_id2, jct['id'])
        del ret_obj['data']['active_worker_threads']
        for srv in ret_obj['data']['servers']:
            del srv['current_requests']
            del srv['operation_state']
            del srv['server_state']
            if ret_obj['data']['stateful_junction'] == 'no':
                del srv['server_uuid']
            del srv['total_requests']
        jct['details'] = ret_obj['data']

    return tools.json_compare(ret_obj1, ret_obj2, deleted_keys=['active_worker_threads', 'servers/current_requests',
                                                                'servers/operation_state', 'servers/server_state',
                                                                'servers/server_uuid', 'servers/total_requests'])
